#    Bitsquid Blender Tools
#    Copyright (C) 2021  Lucas Schwiderski
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


import zlib
import json
import bpy
import mathutils


def parse_sjson(file_path, skip_editor_data=True):
    """
        Translate Bitsquid .bsi SJSON data to plain JSON,
        then parse into a python dictionary.

        Taken from `bsi_import` in the Vermintide 2 SDK, but slightly
        modified to fix some issues and improve readability.
    """
    return_dict = {}
    try:
        with open(file_path, 'rb') as f:
            data = f.read()

            if data[:4] == 'bsiz':
                data = zlib.decompress(data[8:])

            data = data.decode("utf-8")

            file_lines = ['{\n']
            inside_list = False
            check_editor_data = 1 if skip_editor_data else 0

            for line in data.splitlines():
                if check_editor_data:
                    if check_editor_data > 1:
                        if line[0] == '}':
                            check_editor_data = 0

                        continue
                    elif line[:18] == 'editor_metadata = ':
                        check_editor_data = 2
                        continue

                if not line.strip():
                    continue

                # Strip trailing whitespace,
                # including line break and carriage return
                line = line.rstrip()
                if line[-1] in ['\n', '\r']:
                    line = line[:-1]

                if ' = ' in line:
                    line_parts = line.split(' = ')

                    if '[ ' in line_parts[-1] and ' ]' in line_parts[-1]:
                        new_end = ''
                        end_parts = line_parts[-1].split(' ]')
                        short_len = len(end_parts) - 1

                        for i, end_part in enumerate(end_parts):
                            if not end_part:
                                if i < short_len:
                                    new_end += ' ]'
                                continue

                            if '[ ' not in end_part:
                                new_end += end_part + ' ]'
                                continue

                            sub_part_pre = end_part.rpartition('[ ')
                            new_end = ''.join((
                                new_end,
                                sub_part_pre[0],
                                '[ ',
                                ', '.join(sub_part_pre[-1].split(' ')),
                                ' ]'
                            ))

                            if i < short_len and end_parts[i + 1] and end_parts[i + 1][:2] == ' [':
                                new_end += ','

                        line_parts[-1] = new_end

                    # Handle indentation
                    if '\t' in line_parts[0]:
                        line_start = line_parts[0].rpartition('\t')
                    else:
                        tab_len = len(line_parts[0]) - len(line_parts[0].lstrip())

                        if tab_len > 4:
                            line_start = [line_parts[0][:tab_len - 4], '' , line_parts[0].lstrip()]
                        elif tab_len > 0:
                            line_start = ['', '', line_parts[0].lstrip()]
                        else:
                            line_start = line_parts[0].rpartition('\t')

                    if line_start[-1][0] == '"':
                        new_line = ''.join((
                            ''.join(line_start[:-1]),
                            line_start[-1],
                            ': ',
                            ''.join(line_parts[1:])
                        ))
                    else:
                        new_line = ''.join((
                            ''.join(line_start[:-1]),
                            '"',
                            line_start[-1],
                            '": ',
                            ''.join(line_parts[1:])
                        ))

                    if not line_parts[-1][-1] == ',':
                        new_line += ','
                elif ']' in line or '}' in line:
                    new_line = line + ','
                    if '} {' in new_line:
                        new_line = new_line.replace('} {', '}, {')
                    if file_lines[-1][-2] == ',':
                        file_lines[-1] = file_lines[-1][:-2] + '\n'
                elif inside_list and ']' not in line:
                    tab_len = len(line) - len(line.lstrip())
                    new_line = line[:tab_len] + ', '.join([x for x in line[tab_len:].split(' ') if x])
                else:
                    new_line = line

                if new_line[-2:] in ['[,', '{,'] or new_line[-3:] in ['[ ,', '{ ,']:
                    new_line = new_line[:-1]

                if inside_list:
                    if ']' in line:
                        inside_list = False
                    elif not new_line[-1] in ['[', '{', ',']:
                        new_line += ','
                elif '[' in line:
                    inside_list = True

                file_lines.append(''.join(('\t', new_line.lstrip(), '\n')))

                # To save memory...
                if len(file_lines) > 100000:
                    file_lines = [''.join(file_lines)]

            if file_lines[-1][-2] == ',':
                file_lines[-1] = file_lines[-1][:-2] + '\n'

            file_lines.append('}\n')
            return_dict = json.loads(''.join(file_lines))
    except ValueError:
        print(file_path.replace('\\', '/') + ': SJSON file contains a syntax error')
    return return_dict


def find(arr, f):
    """
    Find a value in a list by executing `f`
    on each item until it returns `true`.
    """
    for key, val in arr.items():
        if f(val, key):
            return val, key


def create_object(self, context, name, node_data, geometries):
    """
    Create a Blender object from a BSI node definition
    and additional data from the file.
    """
    print("[create_object]", name, node_data)
    # A list of tuples, where each tuple represents a Vector3(x, y, z).
    vertices = []
    # A list of tuples, where each tuple contains three indices into `vertices`.
    # Those three indices define the vertices that make up the face.
    faces = []

    tri_count = int(geometries["indices"]["size"]) / 3

    for i, index_stream in enumerate(geometries["indices"]["streams"]):
        data_stream = geometries["streams"][i]
        stride = data_stream["stride"] / 4

        for channel in data_stream['channels']:
            stream_data = data_stream["data"]
            if channel['name'] == 'POSITION':
                # NOTE: Do we need to handle `stride != 3`?
                # Since the value seems to be fixed per stream, a higher
                # stride would only be possible for objects that can be built
                # entirely from quads, which is very uncommon.
                if stride != 3:
                    raise RuntimeError("stride != 3 cannot be handled")

                # Get vertex positions.
                # Iterate over data in sets of three values that represent
                # `x`, `y` and `z`.
                for j in range(0, len(stream_data), 3):
                    vertices.append((
                        stream_data[j],
                        stream_data[j + 1],
                        stream_data[j + 2],
                    ))

                # Get face definitions. Values are vertex indices.
                # Iteration works just like vertices above.
                for j in range(0, len(index_stream), 3):
                    faces.append((
                        index_stream[j],
                        index_stream[j + 1],
                        index_stream[j + 2],
                    ))
            else:
                # TODO: Implement other channel types
                self.report(
                    {'WARNING'},
                    "Unknown channel type: {}".format(channel["name"])
                )

    mesh = bpy.data.meshes.new(name)
    mesh.from_pydata(vertices, [], faces)

    return mesh


def import_node(self, context, name, node_data, global_data):
    """Import a BSI node. Recurses into child nodes."""
    print("[import_node]", name, node_data)
    if "geometries" in node_data:
        geometry_name = node_data["geometries"][0]
        print("[import_node] Building geometry '{}' for object '{}'".format(geometry_name, name))
        geometries = global_data["geometries"][geometry_name]
        mesh = create_object(self, context, name, node_data, geometries)
        obj = bpy.data.objects.new(mesh.name, mesh)
    else:
        print("[import_node] Adding empty for '{}'".format(name))
        obj = bpy.data.objects.new(name, None)

    # Needs to happen before scaling can be applied.
    # TODO: Check if above is true, was seen on Stackoverflow.
    obj.matrix_world = mathutils.Matrix()
    # TODO: Apply tranformation matrix in `node_data["local"]`

    if "children" in node_data:
        for child_name, child_data in node_data["children"].items():
            if child_data["parent"] != name:
                raise RuntimeError(
                    "Assigned parent '%s' doesn't match actual parent node '%s'"
                    % (child_data["parent"], name)
                )

            child_obj = import_node(
                self,
                context,
                child_name,
                child_data,
                global_data,
            )
            child_obj.parent = obj

    # Make sure all objects are linked to the current collection.
    # Otherwise they won't show up in the outliner.
    collection = context.collection
    collection.objects.link(obj)
    return obj


def load(self, context, filepath, *, relpath=None):
    global_data = parse_sjson(filepath)

    print(global_data)

    # Nothing to do if there are no nodes
    if "nodes" not in global_data:
        self.report({'WARNING'}, "No nodes to import in {}".format(filepath))
        return {'CANCELLED'}

    view_layer = context.view_layer

    for name, node_data in global_data["nodes"].items():
        obj = import_node(self, context, name, node_data, global_data)

        obj.select_set(True)
        # One of the objects should be set as active.
        # This is the easiest to implement and perfectly fine.
        view_layer.objects.active = obj

    return {'FINISHED'}
